/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 *
 * $Id: gui.c,v 1.5 1998/06/05 19:21:50 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <proplist.h>
#include <gtk/gtk.h>
#include "wtype.h"
#include "proputil.h"
#include "propgroup.h"
#include "propset.h"
#include "gui.h"

static int ReviewGroupChanges( PropertyGroup *group )
{
    int			i, count = 0;
    PropertyEntry	*entry;
    char		*old;
    
    for( i = 0; i < group->numProps; i++ )
    {
	entry = &group->entries[i];
	if( entry->widget ) {
	    proplist_t newval = VALUE_FROM_WIDGET( entry->type, entry->widget);
	    if( entry->value && PLIsEqual( entry->value, newval ))
		continue;
	    count++;
	    if( ! entry->value )
		old = "\"\"";
	    else
		old = PLGetDescription( entry->value );
	    printf( "%s: %s -> %s\n", PLGetString( entry->key ),
		    old, PLGetDescription( newval ));
	}
    }
    return count;
}


static void SaveChanges( PropertySet *set )
{
    int		i, j;
    static int	backed_up = 0;
    
    for( i = 0; i < set->numGroups; i++ )
    {
	PropertyGroup *group = set->groups[i];
	for( j = 0; j < group->numProps; j++ )
	{
	    PropertyEntry *entry = &group->entries[j];
	    if( entry->widget ) {
		proplist_t newval = VALUE_FROM_WIDGET( entry->type,
						       entry->widget);
		if( entry->value && PLIsEqual( entry->value, newval ))
		    continue;
		/* 
		   printf( "Set %s to %s\n", PLGetString( entry->key ),
		   PLGetDescription( newval ));
		*/
		PLInsertDictionaryEntry( set->orig_props, entry->key, newval );
		entry->value = newval;
	    }
	}
    }
    if( backed_up )
    {
	PLSave( set->orig_props, YES );
    }
    else
    {
	PLSaveWithBackup( set->orig_props );
	backed_up++;
    }
}


static void save_callback (GtkWidget *widget, gpointer *data)
{
    PropertySet	*set = (PropertySet *) data;

    SaveChanges( set );
}

static void savemeta_callback (GtkWidget *widget, gpointer *data)
{
    PropertySet	*set = (PropertySet *) data;

    SaveMetaProps( set );
}

static void review_callback (GtkWidget *widget, gpointer *data)
{
    PropertySet	*set = (PropertySet *) data;
    int		i, count = 0;

    for( i = 0; i < set->numGroups; i++ )
    {
	count += ReviewGroupChanges( set->groups[i] );
    }
    if( count )
	printf( "----------------------\n" );
}

static void exit_callback (GtkWidget *widget, gpointer *data)
{
    gtk_main_quit ();
}



static GtkWidget *BuildEditDialogue( PropertySet *set, int do_meta )
{
    GtkWidget	*dialog;
    GtkWidget	*widget;
    GtkWidget	*table;
    char	title[255];	/* XXX */

    /* create a new window */
    dialog = gtk_dialog_new ();

    gtk_signal_connect( GTK_OBJECT( dialog ), "destroy",
			GTK_SIGNAL_FUNC( gtk_widget_destroyed ),
			&dialog );

    gtk_signal_connect( GTK_OBJECT( dialog ), "delete_event",
			GTK_SIGNAL_FUNC( exit_callback ), (gpointer) set );

    sprintf( title, "Malaprop - %s", set->name );
    gtk_window_set_title( GTK_WINDOW( dialog ), title);
    gtk_container_border_width( GTK_CONTAINER( dialog ), 0 );

    /* Review Button */
    widget = gtk_button_new_with_label( "Review Changes" );
    gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->action_area ), 
			widget, TRUE, FALSE, 10);
    gtk_signal_connect( GTK_OBJECT( widget ), "clicked",
			GTK_SIGNAL_FUNC( review_callback ), (gpointer) set);
    gtk_widget_show( widget );

    /* Save Button */
    widget = gtk_button_new_with_label( "Save Changes" );
    GTK_WIDGET_SET_FLAGS( widget, GTK_CAN_DEFAULT );
    gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->action_area ), 
			widget, TRUE, FALSE, 10);
    gtk_widget_grab_default( widget );
    gtk_signal_connect( GTK_OBJECT( widget ), "clicked",
			GTK_SIGNAL_FUNC( save_callback ), (gpointer) set);
    gtk_widget_show( widget );

    /* Save Meta Button */
    if( do_meta )
    {
	widget = gtk_button_new_with_label( "Save Meta" );
	gtk_box_pack_start( GTK_BOX( GTK_DIALOG( dialog )->action_area ), 
			    widget, TRUE, FALSE, 10);
	gtk_signal_connect( GTK_OBJECT( widget ), "clicked",
			    GTK_SIGNAL_FUNC( savemeta_callback ),
			    (gpointer) set);
	gtk_widget_show( widget );
    }
    
    /* Exit button */
    widget = gtk_button_new_with_label( "Exit (discard changes)" );
    gtk_box_pack_start( GTK_BOX( GTK_DIALOG(dialog)->action_area), 
			widget, TRUE, FALSE, 10);
    gtk_signal_connect( GTK_OBJECT( widget ), "clicked",
			GTK_SIGNAL_FUNC( exit_callback ), (gpointer) set);
    gtk_widget_show( widget );

    /* Table with actual property widgets */
    table = PropertySetWidgetsBuild( set );
    gtk_box_pack_start( GTK_BOX( GTK_DIALOG(dialog)->vbox), table,
			TRUE, TRUE, 0);
    gtk_widget_show( table );
    return dialog;
}

void MakePropertySetEditor( PropertySet *set, int do_meta )
{
    GtkWidget		*dialog;
    
    dialog = BuildEditDialogue( set, do_meta );
    gtk_widget_show( dialog );
}

