/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: malaprop.c,v 1.5 1998/06/05 19:21:50 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <ctype.h>
#include <pwd.h>
#include <proplist.h>
#include <gtk/gtk.h>
#include "wtype.h"
#include "proputil.h"
#include "propgroup.h"
#include "propset.h"
#include "gui.h"

#define MAXPATH		256	/* XXX */

#ifndef TRUE
#define TRUE 1
#define FALSE 0
#endif

/* XXX: libPL doesn't give us access to tell if a property was changed, so we
 * keep track ourselves (ugh)
 */
int metaPropsChanged = 0;

const char *DomainUserPath( const char *name )
{
    char		path[MAXPATH];
    char		*value;
    struct passwd	*pwd;

    value = getenv( "GNUSTEP_USER_ROOT" );
    if( value ) {
	strcpy( path, value );
    } else {
	if( ( value = getenv( "HOME" )) != NULL ) {
	    strcpy( path, value );
	}
	else if( (( pwd = getpwuid(getuid())) != NULL ) &&
		 ( pwd->pw_dir != NULL )) {
	    strcpy( path, pwd->pw_dir );
	}
	else {
	    return NULL;
	}
	strcat(path, "/GNUstep");
    }
    
    strcat(path, "/Defaults/");
    strcat(path, name);

    return strdup(path);
}

const char *MetaDomainName( const char *name )
{
    char	mname[MAXPATH];	/* XXX */

    sprintf( mname, "Malaprop-%s", name );
    return strdup( mname );
}

proplist_t DomainLoad( const char *name )
{
    const char	*userpath = DomainUserPath( name );
    proplist_t	list = NULL;

    if( ! userpath )
	return NULL;

    list = PLGetProplistWithPath( userpath );
    return list;
}


char *InferName( char *oldname )
{
    char newname[128];		/* XXX */
    char *s = oldname, *t = newname;
    int first;

    for( first = 1; *s; ) {
	if( isupper( *s ) && !first ) {
	    *t++ = ' ';
	}
	*t++ = *s++;
	first = 0;    
    }
    *t = '\0';
    return strdup( newname );
}




int InferMeta( proplist_t meta, proplist_t key, proplist_t value )
{
    static int	count = 0;
    char	*s, buffer[16];
    WidgetType	*type;
    
    if( ! PLGetDict( meta, "Name" ))
    {
	s = InferName( PLGetString(key) );
	PLDictAdd2( meta, "Name", s );
	free( s );
	metaPropsChanged++;
    }
    if( ! PLGetDict( meta, "Group" ))
    {
	/* Put it into an unknown group */
	count++;
	sprintf( buffer, "Unknown %d", (count / 10 ) + 1 );
	PLDictAdd2( meta, "Group", buffer );
	sprintf( buffer, "%d", count % 10 );
	PLDictAdd2( meta, "Order", buffer );
	metaPropsChanged++;
    }
    if( ! PLGetDictionaryEntry( meta, PLMakeString( "Type" )))
    {
	type = WidgetTypeInfer( value );
	if( type )
	    PLDictAdd1( meta, "Type", type->pname);
	else
	    PLDictAdd2( meta, "Type",  "Unknown" );
	metaPropsChanged++;
    }
    if( ! PLGetDict( meta, "Description" ))
    {
	PLDictAdd2( meta, "Description",  "None" );
	metaPropsChanged++;
    }
    if( ! PLGetDict( meta, "Help" ))
    {
	PLDictAdd2( meta, "Help",  "None" );
	metaPropsChanged++;
    }
    return TRUE;
}

WidgetType *MetaType( proplist_t meta )
{
    proplist_t	val;
    WidgetType	*w;

    val = PLGetDict( meta, "Type" );
    if( ! val )
	return NULL;

    w = WidgetTypeByName( PLGetString( val ));
    if( ! w )
    {
	printf( "Warning: Unknown type %s\n", PLGetString( val ));
    }
    return w;
}

int LoadTypes( proplist_t typeprops )
{
    proplist_t	keys, k, typeinfo;
    char	*newtype, *basetype;
    int		i, numtypes;
    
    if( ! PLIsDictionary( typeprops ))
    {
	fprintf( stderr, "Type definitions: Not a dictionary\n" );
	return FALSE;
    }
    keys = PLGetAllDictionaryKeys( typeprops );
    if( ! PLIsArray( keys )) {
	fprintf( stderr, "Type definitions: Key extration error\n" );
	return FALSE;
    }

    numtypes = PLGetNumberOfElements( keys );
    for( i = 0; i < numtypes; i++ )
    {
	k = PLGetArrayElement( keys, i );
	newtype = PLGetString( k );
	typeinfo = PLGetDictionaryEntry( typeprops, k );

	if( ! PLIsDictionary( typeinfo ))
	{
	    fprintf( stderr, "Type definition %s is not a dictionary\n",
		     newtype );
	    return FALSE;
	}
	basetype = PLGetDictString( typeinfo, "Type" );
	if( ! newtype )
	{
	    fprintf( stderr, "No base type defined for type %s\n",
		     newtype );
	    return FALSE;
	}
	if( ! WidgetTypeDerive( basetype, newtype, typeinfo ))
	{
	    fprintf( stderr, "Unable to derive type %s from %s\n",
		     newtype, basetype );
	    return FALSE;
	}
    }
    return TRUE;
}

static int BuildDefaults( PropertySet *set, proplist_t metaprops,
			  proplist_t grouprops  )
{
    proplist_t		keys, k, meta, deflt, groupname;
    int			i, numkeys;
    PropertyGroup	*group;
    PropertyEntry	*entry;
    WidgetType		*wtype;
    
    keys = PLGetAllDictionaryKeys( metaprops );
    numkeys = PLGetNumberOfElements( keys );
    for( i = 0; i < numkeys; i++ ) {

	k = PLGetArrayElement( keys, i );
	
	meta = PLGetDictionaryEntry( metaprops, k );
	deflt = PLGetDict( meta, "Default" );
	if( deflt )
	{
	    /* printf( "Adding default entry for %s\n", PLGetString( k )); */
	    groupname = PLGetDict( meta, "Group" );
	    if( ! groupname )
	    {
		fprintf( stderr, "Default for %s has no group\n",
			 PLGetString( k ));
		return FALSE;
	    }
	    wtype = MetaType( meta );
	    if( ! wtype )
	    {
		fprintf( stderr, "Default for %s has unknown type\n",
			 PLGetString( k ));
		return FALSE;
	    }
	    /* printf( "Type is %s (0x%x)\n", wtype->name, (int) wtype ); */
	    group = PropertySetFindGroup( set, groupname, grouprops );
	    entry = PropertyGroupNewEntry( group );
	    entry->key = k;
	    entry->meta = meta;
	    entry->type = wtype;
	    entry->value = NULL;
	    entry->deflt = deflt;
	}
    }

    return TRUE;
}


int BuildProps( PropertySet *set, proplist_t props, proplist_t malaprops )
{
    proplist_t		metaprops, grouprops, keys, meta, k, v, groupname;
    proplist_t		typeprops;
    int			i, numkeys;
    PropertyGroup	*group;
    PropertyEntry	*entry;

    if( ! PLIsDictionary( props )) {
	fprintf( stderr, "Not a dictionary\n" );
	return FALSE;
    }
    keys = PLGetAllDictionaryKeys( props );
    if( ! PLIsArray( keys )) {
	fprintf( stderr, "Key extraction failed\n" );
	return FALSE;
    }

    /* Find meta properties dictionary inside malaprops */
    metaprops = PLGetDictionaryEntry( malaprops,
				      PLMakeString( "MetaProperties") );
    if( ! metaprops )
    {
	fprintf( stderr, "Unable to find metaproperties dictionary\n" );
	return FALSE;
    }
    /* Find group properties dictionary */
    grouprops = PLGetDictionaryEntry( malaprops,
				      PLMakeString( "Groups") );
    if( ! grouprops )
    {
	fprintf( stderr, "Unable to find groups dictionary\n" );
	return FALSE;
    }

    /* Find type definitions dictionary */
    typeprops = PLGetDictionaryEntry( malaprops,
				      PLMakeString( "Types") );
    if( typeprops )
    {
	if( ! LoadTypes( typeprops ))
	{
	    fprintf( stderr, "Type load failure\n" );
	    return FALSE;
	}
    }

    if ( ! BuildDefaults( set, metaprops, grouprops ))
    {
	fprintf( stderr, "Unable to build defaults\n" );
	return FALSE;
    }

    numkeys = PLGetNumberOfElements( keys );
    for( i = 0; i < numkeys; i++ ) {

	k = PLGetArrayElement( keys, i );
	PLRetain( k );
	
	v = PLGetDictionaryEntry( props, k );
	PLRetain( v );

	meta = PLGetDictionaryEntry( metaprops, k );
	if( ! meta ) {
	    meta = PLMakeDictionaryFromEntries( NULL, NULL );
	    PLInsertDictionaryEntry( metaprops, k, meta );
	    metaPropsChanged++;
	}
	if( !meta ) {
	    fprintf( stderr, "Metadata failure\n" );
	    return FALSE;
	}
	InferMeta( meta, k, v );

	groupname = PLGetDict( meta, "Group" );
	if( ! groupname )
	{
	    fprintf( stderr, "Can't work out group name!\n" );
	    return FALSE;
	}
	group = PropertySetFindGroup( set, groupname, grouprops );
	if( ! ( entry = PropertyGroupFindEntry( group, k )))
	{
	    /* printf( "Creating new entry for %s\n", PLGetString( k )); */
	    entry = PropertyGroupNewEntry( group );
	    if( ! entry )
	    {
		fprintf( stderr, "Unable to create new property entry!\n" );
		return FALSE;
	    }
	    memset( entry, 0, sizeof( PropertyEntry ));
	    entry->key = k;
	    entry->meta = meta;
	    entry->type = MetaType( meta );
	}
	entry->value = v;
    }
    PropertySetSort( set );
    PLRelease( keys );
    return TRUE;
}

int SaveMetaProps( PropertySet *set )
{
    static int		backed_up = 0;
    const char		*metadomain = MetaDomainName( set->name );
    int			i, j;
    proplist_t		malaprops, metaprops, grouprops;
    PropertyGroup	*group;
    PropertyEntry	*ent;

#if 0
    malaprops = PLMakeDictionaryFromEntries( NULL, NULL );
    PLSetFilename( malaprops,
		   PLMakeString( (char *) DomainUserPath( metadomain )));
    grouprops = PLMakeDictionaryFromEntries( NULL, NULL );
    metaprops = PLMakeDictionaryFromEntries( NULL, NULL );
    PLDictAdd1( malaprops, "Groups", grouprops );
    PLDictAdd1( malaprops, "MetaProperties", metaprops );
#endif
    /* Find meta properties dictionary inside malaprops */
    metaprops = PLGetDictionaryEntry( set->orig_metaprops,
				      PLMakeString( "MetaProperties") );
    if( ! metaprops )
    {
	fprintf( stderr, "Unable to find metaproperties dictionary\n" );
	return FALSE;
    }

    /* Find group properties dictionary too */
    grouprops = PLGetDictionaryEntry( set->orig_metaprops,
				      PLMakeString( "Groups") );
    if( ! grouprops )
    {
	fprintf( stderr, "Unable to find groups dictionary\n" );
	return FALSE;
    }

    printf( "Saving." );
    fflush( stdout );
    for( i = 0; i < set->numGroups; i++ )
    {
	group = set->groups[i];

	PLInsertDictionaryEntry( grouprops, group->name, group->gprops );
	for( j = 0; j < group->numProps; j++ )
	{
	    ent = &group->entries[j];

	    PLGetDescription( set->orig_metaprops );
	    /*
	      printf( "%d %d E: %s => %s\n", i, j,
	      PLGetDescription( ent->key ), PLGetDescription( ent->meta ));
	    */
	    PLInsertDictionaryEntry( metaprops, ent->key, ent->meta );
	    printf( "." );
	    fflush( stdout );
	}
    }
    

    if( backed_up )
    {
	PLSave( set->orig_metaprops, YES );
    }
    else
    {
	PLSaveWithBackup( set->orig_metaprops );
	backed_up++;
    }
    printf( "done\n" );
    return TRUE;
}

proplist_t GetMetaProps( char *domain, int do_default )
{
    const char *metadomain = MetaDomainName( domain );
    proplist_t	malaprops, groups, metaprops;

    /* Load domain if possible */
    malaprops = DomainLoad( metadomain );
    if( !malaprops && do_default ) {
	/* No: create an empty set: A dictionary with 2 entries, one is
	   a dictionary for group information, the other a  group for
	   meta properties */
	fprintf( stderr, "Creating meta properties dictionary\n" );
	malaprops = PLMakeDictionaryFromEntries( NULL, NULL );
	PLSetFilename( malaprops,
		       PLMakeString( (char *)DomainUserPath( metadomain )));
	groups = PLMakeDictionaryFromEntries( NULL, NULL );
	metaprops = PLMakeDictionaryFromEntries( NULL, NULL );
	PLDictAdd1( malaprops, "Groups", groups );
	PLDictAdd1( malaprops, "MetaProperties", metaprops );
	metaPropsChanged++;
    }
    if( metadomain )
	free( (char *) metadomain );
    return malaprops;
}


int MergeDefaults( PropertySet *set, char *filename )
{
    proplist_t	defaults, metaprops, keys, k, meta, curr_def, new_def;
    int		i, numkeys;

    /* Load defaults file */
    defaults = PLGetProplistWithPath( filename );
    if( ! defaults ) {
	fprintf( stderr, "Unable to load defaults from %s\n", filename );
	return FALSE;
    }
    /* Find our metaproperties in the set */
    metaprops = PLGetDict( set->orig_metaprops, "MetaProperties");
    if( ! metaprops )
    {
	fprintf( stderr, "Unable to find metaproperties dictionary\n" );
	return FALSE;
    }
    /* Iterate our metaproperties */
    keys = PLGetAllDictionaryKeys( metaprops );
    numkeys = PLGetNumberOfElements( keys );
    for( i = 0; i < numkeys; i++ ) {
	k = PLGetArrayElement( keys, i );
	/* Look up in defaults list */
	new_def = PLGetDictionaryEntry( defaults, k );
	if( ! new_def )
	    continue;		/* No new default value */

	/* Find existing default value */
	meta = PLGetDictionaryEntry( metaprops, k );
	curr_def = PLGetDict( meta, "Default" );
	if( curr_def && PLIsEqual( new_def, curr_def ))
	    continue;		/* No change  */
	printf( "Would set default of %s to %s\n", PLGetString( k ),
		PLGetDescription( new_def ));
	PLDictAdd1( meta, "Default", new_def );
	metaPropsChanged++;
    }
    return TRUE;
}

int main (int argc, char *argv[])
{
    proplist_t		props, metaprops;
    PropertySet		*set;
    char		*domain;
    int			do_meta = FALSE;
    int			do_def = FALSE;
    char		*def_file;

    gtk_init (&argc, &argv);

    /* XXX proper arg parsing :) */
    if( argc >= 2 && !strcmp( argv[1], "-meta" ))
    {
	do_meta = TRUE;
	argc--;	argv++;		/* shift args */
    }
    if( argc >= 2 && !strcmp( argv[1], "-builddef" ))
    {
	do_def = TRUE;
	argc--;	argv++;		/* shift args */
	if( argc < 2 )
	{
	    printf( "No defaults file supplied\n" );
	    exit( 1 );
	}
	def_file = argv[1];
	argc--;	argv++;		/* shift args */
    }


    if( argc != 2 )
    {
	printf( "Usage: Malaprop <domain>\n" );
	exit( 1 );
    }
    domain = argv[1];
    
    if( (props = DomainLoad( domain )) == NULL ) {
	fprintf( stderr, "Warning: Unable to load properties for domain %s\n",
		 domain );
    }
    if( ( metaprops = GetMetaProps( domain, (props != NULL) )) == NULL ) {
	fprintf( stderr, "Metaproperty domain load error\n" );
	exit( 1 );
    }

    if( !props )
    {
	fprintf( stderr, "Creating empty property list for %s\n", domain );
	props = PLMakeDictionaryFromEntries( NULL, NULL );
	PLSetFilename( props, PLMakeString( (char *)DomainUserPath( domain )));
	PLSave( props, YES );
    }
    set = PropertySetCreate( domain, props, metaprops );
    if( !BuildProps( set, props, metaprops )) {
	fprintf( stderr, "BuildProps error\n" );
	exit( 1 );
    }
    if( do_def )
    {
	printf( "Merging defaults from %s\n", def_file );
	if( !MergeDefaults( set, def_file ))
	{
	    fprintf( stderr, "Defaults merge failed\n" );
	    exit( 0 );
	}
    }
    if( metaPropsChanged )
    {
	printf( "Meta properties changed...\n" );
	SaveMetaProps( set );
    }
    
    MakePropertySetEditor( set, do_meta );
    gtk_main();

    return 0;
}

 
