/*
 * Copyright (c) 1998 Peter Bentley (pete@sorted.org)
 * 
 * Permission to use, copy, modify, distribute, and sell this software
 * and its documentation for any purpose is hereby granted without fee,
 * provided that the above copyright notice and this permission notice
 * appear in all copies of the software and related documentation
 * 
 * THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY
 * WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.
 * 
 * IN NO EVENT SHALL PETER BENTLEY BE LIABLE FOR ANY SPECIAL, INCIDENTAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT
 * ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,
 * ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS
 * SOFTWARE.
 * 
 * $Id: propup.c,v 1.3 1998/05/02 19:53:23 pete Exp $
 * 
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <proplist.h>

#ifndef TRUE
#define TRUE (1)
#define FALSE (0)
#endif

typedef enum
{
    PROP_UNKNOWN,
    PROP_KEY_COMPARE,
    PROP_KEY_MERGE,
    PROP_VALUE_COMPARE,
} pmode;


/* Pinched from WindowMaker for case-insensitive comparisons */
static BOOL
StringCompareHook(proplist_t pl1, proplist_t pl2)
{
    char *str1, *str2;

    str1 = PLGetString(pl1);
    str2 = PLGetString(pl2);

    if (strcasecmp(str1, str2)==0)
	return YES;
    else
	return NO;
}

/* Maybe this one should go inside libPL and work on internal representations
 * of properties :)
 */
static int str_compare( const void *v1, const void *v2 )
{
    char **p1 = ( char ** ) v1, **p2 = ( char ** ) v2;
    
    return strcasecmp(*p1, *p2 );
}

int PLSortStringArray( proplist_t array )
{
    proplist_t	prop;
    char	*s, **strings;
    int		i, size;

    size = PLGetNumberOfElements( array );
    if( ( strings = malloc( size * sizeof( char * ))) == NULL )
	return FALSE;
    
    memset( strings, 0, size * sizeof( char * ));
    /* Pull pointers to copies of strings out of the array */
    for( i = 0; i < size; i++ )
    {
	prop = PLGetArrayElement( array, i );
	if( ! PLIsString( prop ))
	    break;
	s = strdup( PLGetString( prop ));
	if( s == NULL )
	    break;
	strings[i] = s;
    }

    /* Only bother sorting if the above succeeded */
    if( i == size )
    {
	qsort( strings, size, sizeof( char * ), str_compare );
	/* Empty the array */
	for( i = 0; i < size; i++ )
	    PLRemoveArrayElement( array, 0 );
	/* Stuff sorted values back into the array */
	for( i = 0; i < size; i++ ) 
	    PLAppendArrayElement( array, PLMakeString( strings[i] ));
    }

    /* Free up our memory */
    for( i = 0; i < size; i++ )
    {
	if( strings[i] )
	    free( strings[i] );
    }
    free( strings );
    return TRUE;
}

int DoKeyMerge( proplist_t p1, proplist_t p2, int merge )
{
    proplist_t	key, value, newvalue, allkeys;
    int		i, numkeys, differences = 0;;


    /* Get all the keys from the 'new' dictionary */
    allkeys = PLGetAllDictionaryKeys( p2 );
    if( ! PLIsArray( allkeys )) {
	fprintf( stderr, "Key extraction failed\n" );
	return -1;
    }

    PLSortStringArray( allkeys );	/* Alphabetical order is nice */
    numkeys = PLGetNumberOfElements( allkeys );
    for( i = 0; i < numkeys; i++ )
    {
	key = PLGetArrayElement( allkeys, i ); 	/* Get the key */
	value = PLGetDictionaryEntry( p1, key ); /* Look up in 'old' dict */

	if( ! value )
	{
	    differences++;
	    value = PLGetDictionaryEntry( p2, key ); /* Look it up in new... */
	    /* Not found */
	    printf( "Different key %s (value %s)\n",
		    PLGetString( key ), PLGetDescription( value ));
	    if( merge )
	    {
		newvalue = PLDeepCopy( value );
		PLInsertDictionaryEntry( p1, key, newvalue );
	    }
	}
    }
    return differences;
}

int KeyCompare( proplist_t p1, proplist_t p2 )
{
    int ch;
    
    printf( "Comparing...\n" );
    ch =  DoKeyMerge( p1, p2, FALSE );
    if( ch < 0 ) {
	fprintf( stderr, "Comparison failed!\n" );
	return ch;
    }
    if( ch > 0 )
	return ch;

    /* ch == 0 */
    printf( "No difference? Comparing the other way...\n" );
    ch = DoKeyMerge( p2, p1, FALSE );
    if( ch < 0 ) {
	fprintf( stderr, "Comparison failed!\n" );
	return ch;
    }
    if( ch == 0 )
	return ch;
    printf( "Found %d changes the other way...check command line args\n", ch );
    return 0;
}

void KeyMerge( proplist_t p1, proplist_t p2 )
{
    int ch = DoKeyMerge( p1, p2, TRUE );

    if( ch > 0 )
    {
	printf( "Saving %d changes in %s...\n", ch,
		PLGetString( PLGetFilename( p1 )));
	PLSave( p1, TRUE );
    }
}

int ValueCompare( proplist_t p1, proplist_t p2 )
{
    proplist_t	key, v1, v2, allkeys;
    int		i, numkeys, differences = 0;;


    printf( "Keys only in in %s:-\n", PLGetString( PLGetFilename( p1 )));
    DoKeyMerge( p2, p1, FALSE );
    printf( "Keys only in in %s:-\n", PLGetString( PLGetFilename( p2 )));
    DoKeyMerge( p1, p2, FALSE );
    
    /* Get all the keys from the 'new' dictionary */
    allkeys = PLGetAllDictionaryKeys( p1 );
    if( ! PLIsArray( allkeys )) {
	fprintf( stderr, "Key extraction failed\n" );
	return -1;
    }

    printf( "Keys with different values\n" );
    PLSortStringArray( allkeys );	/* Alphabetical order is nice */
    numkeys = PLGetNumberOfElements( allkeys );
    for( i = 0; i < numkeys; i++ )
    {
	key = PLGetArrayElement( allkeys, i );
	v1 = PLGetDictionaryEntry( p1, key );
	v2 = PLGetDictionaryEntry( p2, key );

	if( !v1 || ! v2 || PLIsEqual( v1, v2 ))
	    continue;

	differences++;
	printf( "Key %s: %s => %s\n",
		PLGetString( key ),
		PLGetDescription( v2 ),
		PLGetDescription( v1 ));
    }
    return differences;
}



void Usage()
{
    fprintf( stderr, "Usage: propup [-keycompare|-keymerge|-valuecompare] files\n\n" );
    fprintf( stderr, "You must specify exactly 2 files of properties\n" );
    fprintf( stderr, "The first is the 'older' file and the second is\n" );
    fprintf( stderr, "the 'newer' one (or the one you want to merge in)\n" );
    exit( 1 );
}


proplist_t LoadWithPath(char *name )
{
    proplist_t	pl;

    if( (pl = PLGetProplistWithPath( name )) == NULL )
    {
	fprintf( stderr, "Unable to load property list %s\n", name );
	return NULL;
    }
    if( !PLIsDictionary( pl ))
    {
	fprintf( stderr, "Property list %s is not a dictionary\n", name );
	fprintf( stderr, "(support for other property lists may be added later\n");
	return NULL;
    }
    return pl;
}

int main (int argc, char *argv[])
{
    proplist_t	p1, p2;
    pmode	mode = PROP_KEY_COMPARE;


    /* Skip argv[0] */
    argc--; argv++;

    /* Look for options */
    while( **argv == '-' )
    {
	if( ! strcmp( *argv, "-keycompare" ))
	    mode = PROP_KEY_COMPARE;
	else if( !strcmp( *argv, "-keymerge" ))
	    mode = PROP_KEY_MERGE;
	else if( !strcmp( *argv, "-valuecompare" ))
	    mode = PROP_VALUE_COMPARE;
	else
	    Usage();

	argc--; argv++;
    }

    /* Last two arguments are file names of property lists */
    if( argc != 2 )
    {
	Usage();
    }
    p1 = LoadWithPath( argv[0] );
    p2 = LoadWithPath( argv[1] );
    if( !p1 || !p2 )
    {
	exit( 1 );
    }

    PLSetStringCmpHook(StringCompareHook);
    switch( mode )
    {
	case PROP_VALUE_COMPARE:
	    ValueCompare( p1, p2 );
	    break;

	case PROP_KEY_COMPARE:
	    KeyCompare( p1, p2 );
	    break;

	case PROP_KEY_MERGE:
	    KeyMerge( p1, p2 );
	    break;

	default:
	    fprintf( stderr, "Unknown mode...exitting\n" );
	    exit( 1 );
    }
    return 0;
}
